@@ -1,16 +1,20 @@
 """Python futures for Condor clusters."""
 from concurrent import futures
+from itertools import count
 import os
 import sys
 import threading
 import time
+import traceback
+
 from . import condor
 from . import slurm
-from .remote import INFILE_FMT, OUTFILE_FMT
-from .util import random_string, local_filename
+from .util import (
+    random_string, local_filename, INFILE_FMT, OUTFILE_FMT,
+)
 import cloudpickle
 
-__version__ = '0.4'
+__version__ = '0.5'
 
 LOGFILE_FMT = local_filename('cfut.log.%s.txt')
 
@@ -21,6 +25,10 @@ class RemoteException(Exception):
     def __str__(self):
         return '\n' + self.error.strip()
 
+class JobDied(Exception):
+    pass
+
+
 class FileWaitThread(threading.Thread):
     """A thread that polls the filesystem waiting for a list of files to
     be created. When a specified file is created, it invokes a callback.
@@ -30,17 +38,16 @@ class FileWaitThread(threading.Thread):
         associated with the filename of each file that is created.
         ``interval`` specifies the polling rate.
         """
-        threading.Thread.__init__(self)
+        threading.Thread.__init__(self, daemon=True)
         self.callback = callback
         self.interval = interval
         self.waiting = {}
-        self.lock = threading.Lock()
+        self.lock = threading.Lock()  # To protect the .waiting dict
         self.shutdown = False
 
     def stop(self):
         """Stop the thread soon."""
-        with self.lock:
-            self.shutdown = True
+        self.shutdown = True
 
     def wait(self, filename, value):
         """Adds a new filename (and its associated callback value) to
@@ -50,22 +57,32 @@ class FileWaitThread(threading.Thread):
             self.waiting[filename] = value
 
     def run(self):
-        while True:
+        for i in count():
+            if self.shutdown:
+                return
+
             with self.lock:
-                if self.shutdown:
-                    return
-
-                # Poll for each file.
-                for filename in list(self.waiting):
-                    if os.path.exists(filename):
-                        self.callback(self.waiting[filename])
-                        del self.waiting[filename]
-
+                self.check(i)
             time.sleep(self.interval)
 
+    def check(self, i):
+        """Do one check for completed jobs
+
+        The i parameter allows subclasses like SlurmWaitThread to do something
+        on every Nth check.
+        """
+        # Poll for each file.
+        for filename in list(self.waiting):
+            if os.path.exists(filename):
+                self.callback(self.waiting[filename])
+                del self.waiting[filename]
+
+
 class ClusterExecutor(futures.Executor):
     """An abstract base class for executors that run jobs on clusters.
     """
+    wait_thread_cls = FileWaitThread
+
     def __init__(self, debug=False, keep_logs=False):
         os.makedirs(local_filename(), exist_ok=True)
         self.debug = debug
@@ -76,7 +93,7 @@ class ClusterExecutor(futures.Executor):
         self.jobs_empty_cond = threading.Condition(self.jobs_lock)
         self.keep_logs = keep_logs
 
-        self.wait_thread = FileWaitThread(self._completion)
+        self.wait_thread = self.wait_thread_cls(self._completion)
         self.wait_thread.start()
 
     def _start(self, workerid, additional_setup_lines):
@@ -100,18 +117,25 @@ class ClusterExecutor(futures.Executor):
         if self.debug:
             print("job completed: %i" % jobid, file=sys.stderr)
 
-        with open(OUTFILE_FMT % workerid, 'rb') as f:
-            outdata = f.read()
-        success, result = cloudpickle.loads(outdata)
-
-        if success:
-            fut.set_result(result)
+        try:
+            with open(OUTFILE_FMT % workerid, 'rb') as f:
+                outdata = f.read()
+        except FileNotFoundError:
+            fut.set_exception(
+                JobDied(f"Cluster job {jobid} finished without writing a result")
+            )
         else:
-            fut.set_exception(RemoteException(result))
+            success, result = cloudpickle.loads(outdata)
+
+            if success:
+                fut.set_result(result)
+            else:
+                fut.set_exception(RemoteException(result))
+
+            os.unlink(OUTFILE_FMT % workerid)
 
         # Clean up communication files.
         os.unlink(INFILE_FMT % workerid)
-        os.unlink(OUTFILE_FMT % workerid)
 
         self._cleanup(jobid)
 
@@ -150,6 +174,28 @@ class ClusterExecutor(futures.Executor):
         self.wait_thread.stop()
         self.wait_thread.join()
 
+class SlurmWaitThread(FileWaitThread):
+    slurm_poll_interval = 30
+
+    def check(self, i):
+        super().check(i)
+        if i % (self.slurm_poll_interval // self.interval) == 0:
+            try:
+                finished_jobs = slurm.jobs_finished(self.waiting.values())
+            except Exception:
+                # Don't abandon completion checking if jobs_finished errors
+                traceback.print_exc()
+                return
+
+            if not finished_jobs:
+                return
+
+            id_to_filename = {v: k for (k, v) in self.waiting.items()}
+            for finished_id in finished_jobs:
+                self.callback(finished_id)
+                self.waiting.pop(id_to_filename[finished_id])
+
+
 class SlurmExecutor(ClusterExecutor):
     """Futures executor for executing jobs on a Slurm cluster.
 
@@ -157,15 +203,23 @@ class SlurmExecutor(ClusterExecutor):
     passed to sbatch. They may include sbatch options (starting with
     '#SBATCH') and shell commands, e.g. to set environment variables.
     """
-    def __init__(self, debug=False, keep_logs=False, additional_setup_lines=()):
+    wait_thread_cls = SlurmWaitThread
+
+    def __init__(self, debug=False, keep_logs=False, additional_setup_lines=(),
+                 additional_import_paths=()):
         super().__init__(debug, keep_logs)
         self.additional_setup_lines = additional_setup_lines
+        self.additional_import_paths = additional_import_paths
 
     def _start(self, workerid, additional_setup_lines):
         if additional_setup_lines is None:
             additional_setup_lines = self.additional_setup_lines
+        if self.additional_import_paths:
+            extra_path = ":".join(self.additional_import_paths)
+        else:
+            extra_path = "!"  # ! for nothing, because '' is valid (CWD)
         return slurm.submit(
-            '{} -m cfut.remote {}'.format(sys.executable, workerid),
+            [sys.executable, '-m', 'cfut.remote', workerid, extra_path],
             additional_setup_lines=additional_setup_lines
         )
 
